from datetime import datetime
from sys import stdout
from typing import List, TextIO

from iqrfpy.objects import SensorData
from iqrfpy.utils.sensor_parser import SensorParser
from iqrfpy.utils.sensor_constants import SensorTypes

__all__ = (
    'StringFormatter',
    'PrintUtils',
    'ListManipulation',
    'convert_bytes_to_std_sensor',
    'get_string_between_strings',
)


class StringFormatter:
    """String formatting utility class."""

    DEFAULT_SPACES = 22
    """Default spaces between"""

    @staticmethod
    def format_time(dt: datetime) -> str:
        """Returns a formatted string representing datetime object.

        Args:
            dt (datetime): The datetime object to format.

        Returns:
            str: The formatted datetime string in the format "dd.mm.yyyy  HH:MM:SS.sss".
        """
        return dt.strftime("%d.%m.%Y  %H:%M:%S.%f")[:-3]

    @staticmethod
    def format_pair(text: str, data, spaces: int = DEFAULT_SPACES) -> str:
        """
        Returns a formatted string and data pair with a specific space between them.

        Args:
            text (str): The text to format.
            data: The data to format.
            spaces (int, optional): Minimum width of the text part. Defaults to 22.

        Returns:
            str: The formatted string with the text and data aligned.
        """
        return f'{text:{spaces}} {data}'


class PrintUtils:
    """Print utility class."""

    @staticmethod
    def print_timestamp_pair(text: str, dt: datetime | None = None, spaces: int = StringFormatter.DEFAULT_SPACES,
                             stream: TextIO = stdout):
        """Prints text with formatted timestamp.

        Args:
            text (str): Text to print.
            dt (datetime, optional): datetime object to format. Defaults to datetime.now().
            spaces (int, optional): Minimum width of the text part. Defaults to 22.
            stream (TextIO, optional): Stream to write to. Defaults to sys.stdout.
        """
        if dt is None:
            dt = datetime.now()
        print(
            StringFormatter.format_pair(
                text=text,
                data=StringFormatter.format_time(dt),
                spaces=spaces
            ),
            file=stream
        )

    @staticmethod
    def print_pair(text: str, data, spaces: int = StringFormatter.DEFAULT_SPACES, stream: TextIO = stdout):
        """Prints a formatted pair of text and data.

        Args:
            text (str): Text to print.
            data: Data to print.
            spaces (int, optional): Minimum width of the text part. Defaults to 22.
            stream (TextIO, optional): Stream to write to. Defaults to sys.stdout.
        """
        print(StringFormatter.format_pair(text=text, data=data, spaces=spaces), file=stream)

    @staticmethod
    def print_underlined(text: str, char: str = '-', upper_line: bool = False, line_break: bool = False):
        """Prints a string underlined with a character of choice.

        Args:
            text (str): The text to be printed and underlined.
            char (str, optional): The character used for underlining. Defaults to '-'.
            upper_line (bool, optional): If True, prints a line above the text. Defaults to False.
            line_break (bool, optional): If True, adds a blank line before the underlined text. Defaults to False.
        """
        if line_break:
            print('')
        if upper_line:
            print(f'{char * len(text)}')
        print(f'{text}\n{char * len(text)}')


class ListManipulation:
    """List manipulation utility class."""

    @staticmethod
    def interval_extract(nlist: List[int]):
        """Converts a list of sequential ascending numbers into intervals.

        Args:
           nlist (List[int]): List of sequential numbers.

        Yields:
           List[int]: Intervals as sublists.

        Example:
            >>> ListManipulation.interval_extract([1, 2, 3, 4, 5, 6, 10, 12])
            [[1, 6], [10], [12]]
        """
        length = len(nlist)
        i = 0
        while i < length:
            low = nlist[i]
            while i < length - 1 and nlist[i] + 1 == nlist[i + 1]:
                i += 1
            high = nlist[i]
            if high - low >= 1:
                yield [low, high]
            else:
                yield [low, ]
            i += 1

    @staticmethod
    def get_intervals_from_list(nlist: List[int]) -> str:
        """
        Converts a list of sequential numbers into a formatted string of intervals.

        The function expects a list of unique numbers in ascending order.

        Args:
            nlist (List[int]): List of sequential numbers.

        Returns:
            str: String representing intervals.

        Example:
            >>> ListManipulation.get_intervals_from_list([1, 2, 3, 4, 5, 6, 10, 12])
            '<1,6>, 10, 12'
        """
        tokens = []
        for x in list(ListManipulation.interval_extract(nlist)):
            # Format this [[1, 6], [10], [12]] to this <1,6>, 10, 12
            if len(x) == 1:
                tokens.append(str(x[0]))
            else:
                tokens.append(f'<{x[0]},{x[1]}>')
        return ', '.join(tokens)

    @staticmethod
    def get_list_from_intervals(intervals: str) -> List[int]:
        """
        Converts a string of intervals into a list of numbers.

        Args:
            intervals (str): String representing intervals.

        Returns:
            List[int]: List of numbers.

        Example:
            >>> ListManipulation.get_list_from_intervals('1, <3,6>, 12')
            [1, 3, 4, 5, 6, 12]
        """
        tokens = []
        interval_list = intervals.split(',')
        i = 0
        while i < len(interval_list):
            interval = interval_list[i].strip()

            if interval.startswith('<'):
                # Extract the range within the angle brackets
                start = int(interval[1:])
                i += 1
                end = int(interval_list[i].strip()[:-1])
                tokens += list(range(start, end + 1))
            else:
                tokens.append(int(interval))
            i += 1
        return tokens


def convert_bytes_to_std_sensor(sensor_type: SensorTypes | int, data: List[int]) -> SensorData:
    """Converts raw sensor data into the SensorData object.

    Args:
        sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity).
        data (List[int]): Data to convert.

    Returns:
        :obj:`SensorData`: SensorData objects containing parsed data.

    Raises:
        ValueError: Raised if passed bytes are shorter than required to process a given sensor type

    Example:
        >>> data = [0x01, 0x02]
        >>> sensor = convert_bytes_to_std_sensor(SensorTypes.TEMPERATURE, data[0:2])
        >>> print(f'{sensor.value} {sensor.unit}')
        '32.0625 ˚C'
    """
    sensors = SensorParser.read_sensors_dpa([sensor_type], data)
    return sensors[0]



"""IQMESH network class."""

from typing import List

from iqrfpy.utils.dpa import (
    COORDINATOR_NADR,
    BROADCAST_ADDR,
    IQMESH_TEMP_ADDR,
    LOCAL_DEVICE_ADDR,
)

from .ntw_device import NtwDevice

__all__ = (
    'IqmeshNtw'
)


class IqmeshNtw(object):
    """A class representing an IQMESH network."""

    __slots__ = 'coordinator', 'bonded_nodes', 'discovered_nodes'

    def __init__(self, coordinator: NtwDevice | None = None, bonded_nodes: List[int] | None = None,
                 discovered_nodes: List[int] | None = None):
        """IQMESH network constructor.

        Args:
            coordinator (NtwDevice, optional): Coordinator device. Defaults to None.
            bonded_nodes (List[int], optional): List of bonded node addresses.
                If unspecified, initialized to [].
            discovered_nodes (List[int], optional): List of discovered node addresses.
                If unspecified, initialized to [].
        """
        self.coordinator: NtwDevice | None = coordinator
        """Coordinator device."""
        self.bonded_nodes: List[int] = bonded_nodes or []
        """List of bonded node addresses."""
        self.discovered_nodes: List[int] = discovered_nodes or []
        """List of discovered node addresses."""

    def get_node_count(self) -> int:
        """Get the number of bonded nodes in the network.

        Returns:
            int: The number of bonded Nodes.
        """
        return len(self.bonded_nodes)

    def get_discovered_count(self) -> int:
        """Get the number of discovered nodes in the network.

        Returns:
            int: The number of discovered Nodes.
        """
        return len(self.discovered_nodes)

    def is_bonded(self, nadr: int) -> bool:
        """Check if a node with a given address is bonded to the network.

        Args:
            nadr (int): The address of the Node to check.

        Returns:
            bool: True if the node is bonded, False otherwise.
        """
        return nadr in self.bonded_nodes

    def validate_nadr(self, nadr: int) -> bool:
        """Validate a node address (NADR) for an IQMESH network.

        This function checks if a given node address (NADR) is valid within an IQMESH network. A valid NADR can
        either be bonded to the network or match certain predefined values such as coordinator, broadcast,
        temporary IQMESH address, or local device address.

        Args:
            nadr (int): The node address to be validated.

        Returns:
            bool: True if the NADR is valid, False otherwise.
        """
        return self.is_bonded(nadr) or nadr in [
            COORDINATOR_NADR,
            BROADCAST_ADDR,
            IQMESH_TEMP_ADDR,
            LOCAL_DEVICE_ADDR
        ]

    def get_routing_time(self, nadr: int) -> float:
        """Calculate the routing time based on the longest timeslot in a given RF mode.

        This function calculates the routing time for a specific node address by taking into account the
        safety time and whether it's a coordinator, local device, broadcast, or unicast. The routing time is
        based on the longest timeslot in the RF mode.

        Args:
            nadr (int): Node address for which the routing time is calculated.

        Returns:
            float: The calculated routing time in seconds.
        """
        safety_time = 0.1  # [s]
        if nadr == COORDINATOR_NADR or nadr == LOCAL_DEVICE_ADDR:
            routing_time = 0.5  # [s]
        else:
            routing_time = (self.get_discovered_count() + 1) * self.coordinator.longest_timeslot
            if nadr != BROADCAST_ADDR and nadr != IQMESH_TEMP_ADDR:
                # It is an unicast, so it will also route back
                routing_time *= 2
        return round(routing_time + safety_time, 2)



"""Application helpers objects, enums and data classes."""

from dataclasses import dataclass
from enum import IntEnum

__all__ = (
    'GatewayParams',
    'TransportType'
)


@dataclass
class GatewayParams:
    """Gateway parameters class.

    This data class holds parameters for communication with IQRF Gateway using the MQTT transport,
    and is used by the IqrfApplication class.
    """

    __slots__ = 'address', 'gw_id', 'mqtt_user', 'mqtt_password', 'mqtt_rsp_time', 'mqtt_client_id'

    def __init__(self, address: str, gw_id: str, mqtt_user: str, mqtt_password: str, mqtt_rsp_time: int,
                 mqtt_client_id: str | None = None):
        """Gateway parameters constructor.

        Args:
            address (str): Gateway address, IP address or FQDN.
            gw_id (str): Gateway ID.
            mqtt_user (str): MQTT broker username.
            mqtt_password (str): MQTT broker password.
            mqtt_rsp_time (int): MQTT response timeout in seconds.
            mqtt_client_id (str, optional): MQTT Client ID. If unspecified, a random client ID is generated.
                Defaults to None.
        """
        self.address: str = address
        """Gateway IP or FQDN."""
        self.gw_id: str = gw_id
        """Gateway ID."""
        self.mqtt_user: str = mqtt_user
        """MQTT broker username."""
        self.mqtt_password: str = mqtt_password
        """MQTT broker password."""
        self.mqtt_rsp_time: int = mqtt_rsp_time
        """MQTT response timeout in seconds."""
        self.mqtt_client_id: str | None = mqtt_client_id
        """MQTT Client ID."""


class TransportType(IntEnum):
    """Transport types enum class."""
    MQTT = 0
    IQRF_IDE = 1

    def __str__(self):
        """String representation of transport type.

        Implementation overrides default str() behavior.

        Returns:
            str: Human-readable transport type string.
        """
        return self.name.replace('_', ' ')



from iqrfpy.objects import ExplorationPerEnumData, OsReadData, OsTrConfData
from iqrfpy.utils.common import Common
from iqrfpy.peripherals.exploration import PeripheralEnumerationResponse
from iqrfpy.peripherals.os import ReadTrConfResponse

__all__ = (
    'NtwDevice'
)


class NtwDevice(object):
    """Represents an IQRF network device and provides methods for managing and retrieving device information."""

    __slots__ = 'mid', 'os_ver', 'os_build', 'tr_type', 'supply_voltage', 'shortest_timeslot', 'longest_timeslot', \
        'tr_config_checksum', 'tr_config', 'tr_rfpgm', 'tr_init_phy', 'rf_band', 'is_thermometer', 'is_eeeprom', \
        'is_il_type', 'dpa_ver', 'hwpid', 'hwpid_ver', 'std_lp_ntw_type', 'lp_rf_mode'

    def __init__(self, data: OsReadData | None = None):
        """Network device constructor.

        Args:
            data (OsReadData, optional): Data to initialize the device instance.
        """
        self.shortest_timeslot = None
        self.longest_timeslot = None
        self.tr_config_checksum = None
        self.tr_config: OsTrConfData | None = None
        self.tr_rfpgm = None
        self.tr_init_phy = None
        self.rf_band = None
        self.is_thermometer = None
        self.is_eeeprom = None
        self.is_il_type = None
        self.dpa_ver = None
        self.hwpid = None
        self.hwpid_ver = None
        self.std_lp_ntw_type = None
        self.lp_rf_mode = None
        if data is None:
            return
        self.mid = f'{data.mid:08X}'
        major = (data.os_version & 0xF0) >> 4
        minor = data.os_version & 0x0F
        # TR type
        self.tr_type = data.tr_mcu_type.tr_series
        # OS version
        self.os_ver = f'{major}.{minor:02d}{data.tr_mcu_type.mcu_type}'
        # OS build
        self.os_build = f'{data.os_build:04X}'
        # Power supply
        self.supply_voltage = round(data.supply_voltage, 2)
        # Timeslot limits
        self.shortest_timeslot = ((data.slot_limits & 0x0F) + 3) * 0.01
        self.longest_timeslot = (((data.slot_limits & 0xF0) >> 4) + 3) * 0.01
        # Peripheral enumeration
        self._parse_peripheral_enum(data.per_enum)

    def _parse_peripheral_enum(self, data: ExplorationPerEnumData):
        """
        Saves and parses peripheral enumeration data to the object.

        Args:
          data (PeripheralEnumerationData): The peripheral enumeration data to be saved.
        """
        self.dpa_ver = data.dpa_version
        self.hwpid = data.hwpid
        self.hwpid_ver = data.hwpid_ver
        self.std_lp_ntw_type = bool(data.flags & 0x04)
        self.lp_rf_mode = bool(data.flags & 0x02)

    def device_ident_str(self) -> str:
        """
        Returns a formatted string representing the TR module information.

        Returns:
            str: A string containing the TR module type, MID, IQRF OS version, and build number.
        """
        s = f'{self.tr_type}, MID: {self.mid}, IQRF OS: {self.os_ver} ({self.os_build})'
        s += f'\nDPA: {Common.dpa_version_to_str(self.dpa_ver)}, HWPID: 0x{self.hwpid:04X} (ver: {int(self.hwpid_ver / 256):02X}.{(self.hwpid_ver & 0x00FF):02X})'

        ntw_type = 'STD+LP' if self.std_lp_ntw_type else 'STD'
        rf_mode = 'LP' if self.lp_rf_mode else 'STD'

        s += f', Ntw. type: {ntw_type}, Rf mode: {rf_mode}'
        return s

    def set_peripheral_enum(self, data: PeripheralEnumerationResponse):
        self._parse_peripheral_enum(data.per_enum_data)

    def set_tr_conf(self, data: ReadTrConfResponse) -> bool:
        """
        Saves and parses TR configuration from Read TR configuration response to the object.

        Args:
            data (ReadTrConfResponse): Read TR configuration response.

        Returns:
            bool: True if the checksum of the TR configuration is valid, False otherwise.
        """
        # Check the checksum from the "Configuration" part of PDATA
        if data.checksum != OsTrConfData.calculate_checksum(data.configuration.to_pdata()):
            return False
        self.tr_config_checksum = data.checksum
        self.tr_config = data.configuration
        self.tr_rfpgm = data.rfpgm
        self.tr_init_phy = data.init_phy.value
        self.rf_band = data.init_phy.rf_band
        self.is_thermometer = data.init_phy.thermometer_present
        self.is_eeeprom = data.init_phy.serial_eeprom_present
        self.is_il_type = data.init_phy.il_type
        return True

    def is_per_enabled(self, per: int) -> bool:
        """
        Checks if a specific peripheral in the TR configuration is enabled.

        Args:
            per (int): The peripheral to check.

        Returns:
            bool: True if the peripheral is enabled, False otherwise.
        """
        return per in self.tr_config.embedded_peripherals

    def tr_conf_str(self):
        """
        Returns a formatted string representing the TR configuration.

        Returns:
            str: A string containing TR configuration parameters.
        """
        s = 'Init PHY:'
        s += f'\n\tRf band: {self.rf_band}, Thermometer: {self.is_thermometer}, EEEPROM: {self.is_eeeprom}, IL: {self.is_il_type}'
        s += '\nRF channels:'
        s += f'\n\tA: {self.tr_config.rf_channel_a}'
        s += f'\n\tB: {self.tr_config.rf_channel_b}'
        s += f'\n\tAlternative DSM: {self.tr_config.alternative_dsm_channel}'
        s += '\nRF:'
        ntw_type = 'STD+LP' if self.tr_config.std_and_lp_network else 'STD'
        s += f'\n\tNetwork type: {ntw_type}'
        s += f'\n\tTX power: {self.tr_config.rf_output_power}'
        s += f'\n\tRX filter: {self.tr_config.rf_signal_filter}'
        s += f'\n\tLP RX timeout: {self.tr_config.lp_rf_timeout}'
        s += '\nEmbedded peripherals (enabled):'

        for x in self.tr_config.embedded_peripherals:
            s += f'\n\t{x.name}'

        s += '\nOther:'
        s += f'\n\tCustom DPA Handler: {self.tr_config.custom_dpa_handler}'
        s += f'\n\tDPA Peer-to-Peer: {self.tr_config.dpa_peer_to_peer}'
        s += f'\n\tUser peer-to-peer: {self.tr_config.user_peer_to_peer}'
        s += f'\n\tLocal FRC: {self.tr_config.local_frc}'
        s += f'\n\tIO Setup: {self.tr_config.io_setup}'
        s += f'\n\tRouting off: {self.tr_config.routing_off}'
        s += f'\n\tStay awake when not bonded: {self.tr_config.stay_awake_when_not_bonded}'
        return s



from datetime import datetime
import logging
import sys
import time
from iqrfpy.messages import (
    IRequest,
    IResponse,
    FrcSendReq,
    FrcSendRsp,
    FrcSendSelectiveReq,
    FrcSendSelectiveRsp,
    FrcExtraResultReq,
    FrcExtraResultRsp,
)
from iqrfpy.utils.common import Common
from iqrfpy.utils.dpa import BROADCAST_ADDR, IQMESH_TEMP_ADDR, ResponseCodes, COORDINATOR_NADR, NODE_NADR_MAX

from .iqmesh_ntw import IqmeshNtw
from .objects import GatewayParams, TransportType
from .utils import StringFormatter, PrintUtils

try:
    import iqrfide
except:
    pass


class IqrfApplication:
    """Application class for interacting with IQRF devices and networks using various transport methods."""

    def __init__(self):

        self._app_start_time = None
        self._running_under_iqrfide = None
        self._transport = None
        self._current_transport = None
        self._mqtt_response_time = None
        self._response_code = None
        self.ntw = IqmeshNtw()

        # Logger settings
        self.default_log_level = logging.DEBUG

        self.logger = logging.getLogger()
        self.logger.setLevel(level=self.default_log_level)
        self.logger.addHandler(hdlr=logging.StreamHandler(stream=sys.stdout))

    @property
    def running_under_iqrfide(self):
        """Property to check if the application is running under IQRF IDE.

        Returns:
            bool: True if running under IQRF IDE, False otherwise.
        """
        return self._running_under_iqrfide

    @property
    def current_transport(self):
        """Property to get the current transport type.

        Returns:
            str: The current transport type (e.g., IQRF IDE, MQTT).
        """
        return self._current_transport

    @property
    def response_code(self):
        """Property to get the last response code received.

        Returns:
            int: The last response code received from IQRF device.
        """
        return self._response_code

    def start(self):
        """Starts the application and logs the start time."""
        self._app_start_time = datetime.now()
        PrintUtils.print_timestamp_pair('### Start ###', self._app_start_time)

    def clean_up(self):
        """Cleans up resources if possible."""
        try:
            self._transport.terminate()
        except:  # TODO: specific exceptions
            pass

    def abort(self, message: str):
        """Aborts the application, logs the message and performs cleanup.

        Args:
            message (str): The error message to log.
        """
        print(f'\n{message}')
        PrintUtils.print_timestamp_pair('### Script aborted ###')
        self.clean_up()
        exit()

    def end(self):
        """Ends the application, logs the end time and duration, and performs cleanup."""
        app_end_time = datetime.now()
        exec_time = str(app_end_time - self._app_start_time)[:-3]
        PrintUtils.print_pair(
            '### End ###',
            f'{StringFormatter.format_time(app_end_time)} (duration: {exec_time})'
        )
        self.clean_up()

    def connect(self, transport_type: TransportType = TransportType.IQRF_IDE, gw_params: GatewayParams | None = None):
        """Connects the application to the specified transport method (e.g., IQRF IDE, MQTT) and logs result.

        Args:
            transport_type (TransportType): The transport method to use (default is IQRF IDE).
            gw_params (dict, optional): IQRF gateway configuration (required for MQTT only).

        Raises:
            ImportError: If the specified transport is not available.
        """
        self._current_transport = transport_type
        # IQRF IDE detection
        try:
            import __main__
            self._running_under_iqrfide = __main__.IQRFIDE
        except:
            self._running_under_iqrfide = False

        if self._current_transport == TransportType.IQRF_IDE:
            # Transport IQRF IDE
            try:
                from iqrfpy.ext.iqrfide_transport import IqrfIdeTransport
            except:
                print('\nThe IQRF_IDE transport can be used in the IQRF IDE only')
                exit()
        else:
            # Transport MQTT
            from iqrfpy.ext.mqtt_transport import MqttTransportParams, MqttTransport
            # Treatment of functions extended in IQRF IDE
            if not self._running_under_iqrfide:
                time.sleep_ide = time.sleep

            if gw_params is None:
                print('MQTT transport requires gw_params.')
                exit()

            # MQTT transport configuration
            gw_id = gw_params.gw_id
            self._mqtt_response_time = gw_params.mqtt_rsp_time
            mqtt_params = MqttTransportParams(
                host=gw_params.address,
                port=1883,
                client_id=gw_params.mqtt_client_id,
                user=gw_params.mqtt_user,
                password=gw_params.mqtt_password,
                request_topic=f'gateway/{gw_id.lower()}/iqrf/requests',
                response_topic=f'gateway/{gw_id.lower()}/iqrf/responses',
                qos=1,
                keepalive=25
            )

        PrintUtils.print_pair('Transport:', self._current_transport)

        try:
            if self._current_transport == TransportType.IQRF_IDE:
                self._transport = IqrfIdeTransport()
                PrintUtils.print_pair('IQRF IDE version:', iqrfide.__version__)
                # Enable logging to the IQRF IDE Terminal Log
                self._transport.iqrfide_logging = True
            else:
                self._transport = MqttTransport(params=mqtt_params, auto_init=True)
        except Exception as e:
            self.abort(f'Connection error: {str(e)}')

        PrintUtils.print_timestamp_pair('Connected:')

    def dpa_send_receive(self, request: IRequest, log_level: int = logging.ERROR) -> IResponse:
        """Sends a DPA request, receives a response, and logs the communication details (according to the specified level).

        Args:
            request (IRequest): The DPA request to send.
            log_level (int, optional): The level of details in the log (default is logging.ERROR).
                logging.ERROR: Only errors reported
                logging.INFO: Timestamps, NADRs and errors reported
                logging.DEBUG: Everything reported

        Returns:
            IResponse: The DPA response received, or None if there was no response or if the response code is not ResponseCodes.OK.
        """
        self._response_code = None
        current_log_level = self.logger.level
        self.logger.setLevel(log_level)
        request_nadr = request.nadr
        dpa_rsp_time = request.dpa_rsp_time
        dev_process_time = request.dev_process_time

        self.logger.info(StringFormatter.format_pair('NADR:', request_nadr))
        self.logger.info(StringFormatter.format_pair('Request sent:', StringFormatter.format_time(datetime.now())))
        # DPA request as bytes
        req_as_bytes = request.to_dpa()
        req_as_bytes = ' '.join("{:02X}".format(req_byte) for req_byte in req_as_bytes)
        self.logger.debug(StringFormatter.format_pair('DPA request:', req_as_bytes))

        self.logger.debug(StringFormatter.format_pair('DPA timeout:', f'{dpa_rsp_time} s'))
        self.logger.debug(StringFormatter.format_pair('DPA process time:', f'{dev_process_time} s'))
        if self._current_transport == TransportType.MQTT:
            mqtt_rsp_time = self._mqtt_response_time

            # Calculate DPA response time if not specified
            if dpa_rsp_time is None:
                dpa_rsp_time = self.ntw.get_routing_time(request_nadr)

            # Set MQTT transport timeout: must be longer than DPA timeout
            mqtt_rsp_time += (dpa_rsp_time + (dev_process_time if dev_process_time is not None else 0))
            self.logger.debug(StringFormatter.format_pair('Transport timeout:', f'{mqtt_rsp_time} s'))

        try:
            # Send/receive according to the transport
            if self._current_transport == TransportType.MQTT:
                received_response = self._transport.send_and_receive(request, mqtt_rsp_time)
            else:
                received_response = self._transport.send_and_receive(request)
        except Exception as e:
            self.logger.error(StringFormatter.format_pair(f'Error (NADR {request_nadr}):', str(e)))
        else:
            # Request to broadcast (0xFF) or temporary (0xFE) address returns no response (received_response = None)
            if (request_nadr == BROADCAST_ADDR) or (request_nadr == IQMESH_TEMP_ADDR):
                self.logger.info(StringFormatter.format_pair('No response:', 'broadcast or temporary address'))
                return received_response

            self.logger.info(
                StringFormatter.format_pair(
                    'Response received:',
                    StringFormatter.format_time(datetime.now())
                )
            )
            self._response_code = received_response.rcode
            # Report when there is an error or always at DEBUG level
            if not (ResponseCodes.is_ok_response(self._response_code)) or self.logger.level == logging.DEBUG:
                self.logger.error(
                    StringFormatter.format_pair(
                        f'Rsp code (NADR {request_nadr}):',
                        f'{self._response_code}: {ResponseCodes.to_string(self._response_code)}'
                    )
                )

            if self._current_transport == TransportType.MQTT:
                self.logger.debug(StringFormatter.format_pair('Message type:', received_response.mtype))
                self.logger.debug(StringFormatter.format_pair('Message ID:', received_response.msgid))
            self.logger.debug(
                StringFormatter.format_pair(
                    'PDATA:',
                    Common.list_to_hex_string(received_response.pdata, separator='.', uppercase=True) if
                    received_response.pdata is not None else 'No PDATA.')
            )

            if not ResponseCodes.is_ok_response(self._response_code):
                received_response = None
            return received_response
        finally:
            self.logger.info('------')
            self.logger.setLevel(current_log_level)

    def frc_send_receive(self, frc_cmd: int, user_data: list[int], selected_nodes: list[int] = None) -> dict:
        """Sends (Selective) FRC and Extra Result requests, receives responses, and logs possible errors.

        Args:
            frc_cmd (int): FRC command to send.
            user_data (list[int]): FRC user data to send.
            selected_nodes (list[int], optional): List of Node addresses to which the FRC should be sent
                selectively (default is None - non-selective FRC).

        Returns:
            dict: A dictionary containing the received response data.
                Keys:
                    - 'frc_complete' (bool): Indicates if the FRC operation is complete
                        (both FRC Send and Extra Result responses are received without error).
                    - 'frc_status' (int): FRC Status code.
                    - 'frc_data' (bytes): FRC data.
                    - 'frc_extra_result' (bytes): FRC extra result data.
        """
        rsp_data = {"frc_complete": False, "frc_status": None, "frc_data": None, "frc_extra_result": None}
        # Send/receive FRC or Selective FRC
        if selected_nodes is None:
            response: FrcSendRsp = self.dpa_send_receive(FrcSendReq(COORDINATOR_NADR, frc_cmd, user_data))
        else:
            response: FrcSendSelectiveRsp = self.dpa_send_receive(
                FrcSendSelectiveReq(COORDINATOR_NADR, frc_cmd, selected_nodes, user_data)
            )

        if response:
            err_txt = None
            frc_status = response.status
            rsp_data["frc_status"] = frc_status
            # Check FRC Status according to the IQRF OS Ref. Guide
            if frc_status > NODE_NADR_MAX:
                if frc_status == 0xFF:
                    err_txt = f'No Node is bonded'
                else:
                    err_txt = f'FRC error'
            else:
                # FRC status OK or No Node responded
                rsp_data["frc_data"] = response.frc_data
                # Send/receive FRC Extra Result
                response: FrcExtraResultRsp = self.dpa_send_receive(FrcExtraResultReq(COORDINATOR_NADR))
                if response:
                    # FRC data is complete
                    rsp_data["frc_extra_result"] = response.frc_data
                    rsp_data["frc_complete"] = True

            if err_txt:
                PrintUtils.print_pair('Error:', f'{err_txt} (status: 0x{frc_status:02X})')
        return rsp_data



"""App helpers.

Basic application logic and examples.

.. include:: ../README.md

.. include:: ../changelog.md
"""

from __future__ import annotations

from .iqmesh_ntw import IqmeshNtw
from .iqrf_application import IqrfApplication
from .ntw_device import NtwDevice
from .objects import GatewayParams, TransportType
from .utils import (
    StringFormatter,
    PrintUtils,
    ListManipulation,
    convert_bytes_to_std_sensor,
)

__all__ = (
    # .iqmesh_ntw
    'IqmeshNtw',
    # .iqrf_application
    'IqrfApplication',
    # .ntw_device
    'NtwDevice',
    # .objects
    'GatewayParams',
    'TransportType',
    # .utils
    'StringFormatter',
    'PrintUtils',
    'ListManipulation',
    'convert_bytes_to_std_sensor',
)



from __future__ import annotations
from setuptools import setup

with open('README.md') as file:
    long_description = file.read()

setup(
    name='iqrfpy-app-helpers',
    description='Diagnostics for iqrfpy',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.1.0-dev1',
    url='https://gitlab.iqrf.org/open-source/iqrf-sdk/iqrfpy/iqrfpy-app-helpers',
    author='Karel Hanák',
    author_email='karel.hanak@iqrf.org',
    license='Apache-2.0',
    keywords=['iqrf'],
    project_urls={
        'Homepage': 'https://gitlab.iqrf.org/open-source/iqrf-sdk/iqrfpy/iqrfpy-app-helpers',
        'Changelog': 'https://gitlab.iqrf.org/open-source/iqrf-sdk/iqrfpy/iqrfpy-app-helpers/-/blob/master/changelog.md',
        'Source code': 'https://gitlab.iqrf.org/open-source/iqrf-sdk/iqrfpy/iqrfpy-app-helpers',
        'Issue tracker': 'https://gitlab.iqrf.org/open-source/iqrf-sdk/iqrfpy/iqrfpy-app-helpers/-/issues',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
    ],
    packages=['iqrfpy.ext.app_helpers'],
    package_dir={
        'iqrfpy.ext.app_helpers': 'app_helpers'
    },
    package_data={
        'iqrfpy.ext.app_helpers': [
            'py.typed'
        ]
    },
    python_requires='>=3.10',
    install_requires=[
        'iqrfpy>=0.2.0',
        'tabulate>=0.9.0',
    ]
)
